from abc import ABCMeta, abstractmethod
from typing import List

from twisted.internet.defer import Deferred

from peek_core_search.tuples.ImportSearchObjectTuple import ImportSearchObjectTuple


class SearchApiABC(metaclass=ABCMeta):

    @abstractmethod
    def importSearchObjects(self, searchObjectsEncodedPayload: bytes) -> Deferred:
        """ Import Search Objects

        This method imports a group of objects into the search.

        :param searchObjectsEncodedPayload: A group/list of objects to import into the search
                The format of the encodedPayload tuples is List[ImportSearchObjectTuple]
                using Payload(tuples=tuples).toEncodedPayload()

        :return: A deferred that fires when the objects are queued for indexing.

        """

    @abstractmethod
    def removeSearchObjects(self, importGroupHashes: List[str]) -> Deferred:
        """ Remove Search Objects

        This method removes a group of objects from the search that have "importGroupHashes"

        :param importGroupHashes: A list of strings matching the "importGroupHash"
            provided when the objects were imported.
        :return: A deferred that fires when the objects are queued for deleting.

        """



import logging

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import \
    ACIProcessorQueueControllerABC, ACIProcessorQueueBlockItem
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import \
    ACIProcessorStatusNotifierABC
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import \
    ACIProcessorQueueTupleABC
from peek_core_search._private.server.client_handlers.ClientSearchObjectChunkUpdateHandler import \
    ClientSearchObjectChunkUpdateHandler
from peek_core_search._private.server.controller.StatusController import \
    StatusController
from peek_core_search._private.storage.EncodedSearchObjectChunk import \
    EncodedSearchObjectChunk
from peek_core_search._private.storage.SearchObject import SearchObject
from peek_core_search._private.storage.SearchObjectCompilerQueue import \
    SearchObjectCompilerQueue

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.searchObjectCompilerQueueStatus = state
        self._adminStatusController.status.searchObjectCompilerQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.searchObjectCompilerQueueProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.searchObjectCompilerQueueLastError = error
        self._adminStatusController.notify()


class SearchObjectChunkCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 1.000

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = SearchObjectCompilerQueue
    _VacuumDeclaratives = (SearchObjectCompilerQueue,
                           SearchObject, EncodedSearchObjectChunk)

    def __init__(self, dbSessionCreator,
                 statusController: StatusController,
                 clientSearchObjectUpdateHandler: ClientSearchObjectChunkUpdateHandler):
        ACIProcessorQueueControllerABC.__init__(self, dbSessionCreator,
                                                _Notifier(statusController))
        self._clientSearchObjectUpdateHandler: ClientSearchObjectChunkUpdateHandler \
            = clientSearchObjectUpdateHandler

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_core_search._private.worker.tasks.SearchObjectChunkCompilerTask import \
            compileSearchObjectChunk

        return compileSearchObjectChunk.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientSearchObjectUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return '''
                 with sq_raw as (
                    SELECT "id", "chunkKey"
                    FROM core_search."SearchObjectCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "chunkKey"
                    FROM sq_raw
                    GROUP BY  "chunkKey"
                    HAVING count("chunkKey") > 1
                )
                DELETE
                FROM core_search."SearchObjectCompilerQueue"
                     USING sq sq1
                WHERE core_search."SearchObjectCompilerQueue"."id" != sq1."minId"
                    AND core_search."SearchObjectCompilerQueue"."id" > %(id)s
                    AND core_search."SearchObjectCompilerQueue"."chunkKey" = sq1."chunkKey"

            ''' % {'id': lastFetchedId, 'limit': dedupeLimit}



"""added object order

Peek Plugin Database Migration Script

Revision ID: 5e2b9afbb499
Revises: 36aeb6eaea5c
Create Date: 2019-08-06 10:19:09.787190

"""

# revision identifiers, used by Alembic.
revision = '5e2b9afbb499'
down_revision = '36aeb6eaea5c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('SearchObjectType', sa.Column('order', sa.Integer(), server_default='0', nullable=False), schema='core_search')


    op.execute(''' UPDATE "core_search"."SearchObjectType" SET "order" = 0 ''')

    op.alter_column('SearchObjectType', 'order',
                    type_=sa.Integer(), nullable=False,
                    schema='core_search')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('SearchObjectType', 'order', schema='core_search')
    # ### end Alembic commands ###


"""initial tables

Peek Plugin Database Migration Script

Revision ID: 30f0d0f925ca
Revises: 
Create Date: 2018-06-09 20:19:50.662811

"""

# revision identifiers, used by Alembic.
revision = '30f0d0f925ca'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('EncodedSearchIndexChunk',
    sa.Column('id', sa.BigInteger(), nullable=False),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.Column('encodedData', sa.LargeBinary(), nullable=False),
    sa.Column('encodedHash', sa.String(), nullable=False),
    sa.Column('lastUpdate', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id', 'chunkKey'),
    schema='core_search'
    )
    op.create_index('idx_EncodedSearchIndex_chunkKey', 'EncodedSearchIndexChunk', ['chunkKey'], unique=True, schema='core_search')
    op.create_table('EncodedSearchObjectChunk',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.Column('encodedData', sa.LargeBinary(), nullable=False),
    sa.Column('encodedHash', sa.String(), nullable=False),
    sa.Column('lastUpdate', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id', 'chunkKey'),
    schema='core_search'
    )
    op.create_index('idx_EncodedSearchObject_chunkKey', 'EncodedSearchObjectChunk', ['chunkKey'], unique=True, schema='core_search')
    op.create_table('SearchIndexCompilerQueue',
    sa.Column('id', sa.Integer(), nullable=False, autoincrement=True),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', 'chunkKey'),
    schema='core_search'
    )
    op.create_table('SearchObjectCompilerQueue',
    sa.Column('id', sa.Integer(), nullable=False, autoincrement=True),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', 'chunkKey'),
    schema='core_search'
    )
    op.create_table('SearchObjectType',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('title', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_ObjType_name', 'SearchObjectType', ['name'], unique=True, schema='core_search')
    op.create_index('idx_ObjType_title', 'SearchObjectType', ['title'], unique=True, schema='core_search')
    op.create_table('SearchProperty',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('order', sa.Integer(), server_default='0', nullable=False),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_SearchProp_name', 'SearchProperty', ['name'], unique=True, schema='core_search')
    op.create_index('idx_SearchProp_title', 'SearchProperty', ['title'], unique=True, schema='core_search')
    op.create_table('Setting',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_table('SearchObject',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('objectTypeId', sa.Integer(), nullable=False),
    sa.Column('key', sa.String(), nullable=False),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.Column('propertiesJson', sa.String(), nullable=True),
    sa.Column('packedJson', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['objectTypeId'], ['core_search.SearchObjectType.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_SearchObject_chunkKey', 'SearchObject', ['chunkKey'], unique=False, schema='core_search')
    op.create_index('idx_SearchObject_key', 'SearchObject', ['key'], unique=True, schema='core_search')
    op.create_index('idx_SearchObject_objectTypeId', 'SearchObject', ['objectTypeId'], unique=False, schema='core_search')
    op.create_table('SettingProperty',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('settingId', sa.Integer(), nullable=False),
    sa.Column('key', sa.String(length=50), nullable=False),
    sa.Column('type', sa.String(length=16), nullable=True),
    sa.Column('int_value', sa.Integer(), nullable=True),
    sa.Column('char_value', sa.String(), nullable=True),
    sa.Column('boolean_value', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['settingId'], ['core_search.Setting.id'], ),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_SettingProperty_settingId', 'SettingProperty', ['settingId'], unique=False, schema='core_search')
    op.create_table('SearchIndex',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('chunkKey', sa.Integer(), nullable=False),
    sa.Column('keyword', sa.String(), nullable=False),
    sa.Column('propertyName', sa.String(), nullable=False),
    sa.Column('objectId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['objectId'], ['core_search.SearchObject.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_SearchIndex_quick_query', 'SearchIndex', ['chunkKey', 'keyword', 'propertyName', 'objectId'], unique=True, schema='core_search')
    op.create_table('SearchObjectRoute',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('objectId', sa.Integer(), nullable=False),
    sa.Column('importGroupHash', sa.String(), nullable=False),
    sa.Column('routeTitle', sa.String(), nullable=False),
    sa.Column('routePath', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['objectId'], ['core_search.SearchObject.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='core_search'
    )
    op.create_index('idx_ObjectRoute_objectId', 'SearchObjectRoute', ['objectId'], unique=False, schema='core_search')
    op.create_index('idx_ObjectRoute_routeTitle_importGroupHash', 'SearchObjectRoute', ['importGroupHash'], unique=False, schema='core_search')
    op.create_index('idx_ObjectRoute_routeTitle_objectId', 'SearchObjectRoute', ['routeTitle', 'objectId'], unique=True, schema='core_search')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_ObjectRoute_routeTitle_objectId', table_name='SearchObjectRoute', schema='core_search')
    op.drop_index('idx_ObjectRoute_routeTitle_importGroupHash', table_name='SearchObjectRoute', schema='core_search')
    op.drop_index('idx_ObjectRoute_objectId', table_name='SearchObjectRoute', schema='core_search')
    op.drop_table('SearchObjectRoute', schema='core_search')
    op.drop_index('idx_SearchIndex_quick_query', table_name='SearchIndex', schema='core_search')
    op.drop_table('SearchIndex', schema='core_search')
    op.drop_index('idx_SettingProperty_settingId', table_name='SettingProperty', schema='core_search')
    op.drop_table('SettingProperty', schema='core_search')
    op.drop_index('idx_SearchObject_objectTypeId', table_name='SearchObject', schema='core_search')
    op.drop_index('idx_SearchObject_key', table_name='SearchObject', schema='core_search')
    op.drop_index('idx_SearchObject_chunkKey', table_name='SearchObject', schema='core_search')
    op.drop_table('SearchObject', schema='core_search')
    op.drop_table('Setting', schema='core_search')
    op.drop_index('idx_SearchProp_title', table_name='SearchProperty', schema='core_search')
    op.drop_index('idx_SearchProp_name', table_name='SearchProperty', schema='core_search')
    op.drop_table('SearchProperty', schema='core_search')
    op.drop_index('idx_SearchProp_title', table_name='SearchObjectType', schema='core_search')
    op.drop_index('idx_SearchProp_name', table_name='SearchObjectType', schema='core_search')
    op.drop_table('SearchObjectType', schema='core_search')
    op.drop_table('SearchObjectCompilerQueue', schema='core_search')
    op.drop_table('SearchIndexCompilerQueue', schema='core_search')
    op.drop_index('idx_EncodedSearchObject_chunkKey', table_name='EncodedSearchObjectChunk', schema='core_search')
    op.drop_table('EncodedSearchObjectChunk', schema='core_search')
    op.drop_index('idx_EncodedSearchIndex_chunkKey', table_name='EncodedSearchIndexChunk', schema='core_search')
    op.drop_table('EncodedSearchIndexChunk', schema='core_search')
    # ### end Alembic commands ###


from peek_plugin_base.storage.DbConnection import DbSessionCreator
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchObservableName
from peek_core_search._private.server.tuple_providers.AdminStatusTupleProvider import \
    AdminStatusTupleProvider
from peek_core_search._private.server.tuple_providers.SearchObjectTypeTupleProvider import \
    SearchObjectTypeTupleProvider
from peek_core_search._private.storage.SearchObjectTypeTuple import \
    SearchObjectTypeTuple
from peek_core_search._private.storage.SearchPropertyTuple import SearchPropertyTuple
from peek_core_search._private.tuples.AdminStatusTuple import AdminStatusTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from .controller.StatusController import StatusController
from .tuple_providers.SearchPropertyTupleProvider import SearchPropertyTupleProvider


def makeTupleDataObservableHandler(dbSessionCreator: DbSessionCreator,
                                   statusController: StatusController):
    """" Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param statusController: The search status controller.
    :param dbSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=searchObservableName,
        additionalFilt=searchFilt)

    # Search Property
    tupleObservable.addTupleProvider(SearchPropertyTuple.tupleName(),
                                     SearchPropertyTupleProvider(dbSessionCreator))

    # Search Object Type
    tupleObservable.addTupleProvider(SearchObjectTypeTuple.tupleName(),
                                     SearchObjectTypeTupleProvider(dbSessionCreator))

    # Admin status tuple
    tupleObservable.addTupleProvider(AdminStatusTuple.tupleName(),
                                     AdminStatusTupleProvider(statusController))

    return tupleObservable



from typing import List, Dict

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search.tuples.ImportSearchObjectRouteTuple import \
    ImportSearchObjectRouteTuple
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class ImportSearchObjectTuple(Tuple):
    """ Import Search Object

    This tuple is used by other plugins to load objects into the search index.

    """
    __tupleType__ = searchTuplePrefix + 'ImportSearchObjectTuple'

    #:  The unique key for this object
    # This key will be indexed as a full keyword, do not include the key in the keywords
    key: str = TupleField()

    #:  The type of this object
    objectType: str = TupleField()

    #:  Full Keywords
    # The keywords to index that allows the user to search by partial keywords
    # The key of the property will match of create a new "SearchProperty"
    fullKeywords: Dict[str, str] = TupleField({})

    #:  Partial Keywords
    # The keywords to index that allows the user to search by partial keywords
    # The key of the property will match of create a new "SearchProperty"
    partialKeywords: Dict[str, str] = TupleField({})

    #:  The color
    routes: List[ImportSearchObjectRouteTuple] = TupleField([])



import logging

from sqlalchemy import Column
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index, ForeignKey

from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class SearchObjectRoute(Tuple, DeclarativeBase):
    """ Search Object Route

    This is like the "Open with"

    """
    __tablename__ = 'SearchObjectRoute'
    __tupleType__ = searchTuplePrefix + 'SearchObjectRouteTable'

    id = Column(Integer, primary_key=True, autoincrement=True)

    #:  The object that this routs is for
    objectId = Column(Integer,
                      ForeignKey('SearchObject.id', ondelete='CASCADE'),
                      nullable=False)

    importGroupHash = Column(String, nullable=False)
    routeTitle = Column(String, nullable=False)
    routePath = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_ObjectRoute_objectId", objectId), # For foreign key
        Index("idx_ObjectRoute_routeTitle_importGroupHash", importGroupHash, unique=False),
        Index("idx_ObjectRoute_routeTitle_objectId", routeTitle, objectId, unique=True),
    )



"""add partial kw

Peek Plugin Database Migration Script

Revision ID: 758f26706069
Revises: 0736e757d1ec
Create Date: 2020-06-01 12:35:51.862282

"""

# revision identifiers, used by Alembic.
revision = '758f26706069'
down_revision = '0736e757d1ec'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('SearchObject', 'propertiesJson',
                    new_column_name='fullKwPropertiesJson',
                    schema='core_search')

    op.add_column('SearchObject',
                  sa.Column('partialKwPropertiesJson', sa.String(), nullable=True),
                  schema='core_search')
    # ### end Alembic commands ###


def downgrade():
    op.alter_column('SearchObject', 'fullKwPropertiesJson',
                    new_column_name='propertiesJson',
                    schema='core_search')

    op.drop_column('SearchObject', 'partialKwPropertiesJson', schema='core_search')



from typing import Optional

from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC

from peek_core_search._private.PluginNames import searchTuplePrefix


@addTupleType
class KeywordAutoCompleteTupleAction(TupleActionABC):
    """ Keyword Auto Complete Tuple

    This tuple represents the details of a search result.
    """
    __tupleType__ = searchTuplePrefix + 'KeywordAutoCompleteTupleAction'

    #:  The property key to restrict to
    objectTypeId: Optional[int] = TupleField()

    #:  The property key to restrict to
    propertyName: Optional[str] = TupleField()

    #:  The search string to search for
    searchString: str = TupleField()





from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchObservableName
from vortex.handler.TupleDataObservableProxyHandler import TupleDataObservableProxyHandler


def makeDeviceTupleDataObservableProxy():
    return TupleDataObservableProxyHandler(observableName=searchObservableName,
                                           proxyToVortexName=peekServerName,
                                           additionalFilt=searchFilt)



import logging
from typing import Callable

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import \
    ACIProcessorQueueControllerABC, ACIProcessorQueueBlockItem
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import \
    ACIProcessorStatusNotifierABC
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import \
    ACIProcessorQueueTupleABC
from peek_core_search._private.server.client_handlers.ClientSearchIndexChunkUpdateHandler import \
    ClientSearchIndexChunkUpdateHandler
from peek_core_search._private.server.controller.StatusController import \
    StatusController
from peek_core_search._private.storage.EncodedSearchIndexChunk import \
    EncodedSearchIndexChunk
from peek_core_search._private.storage.SearchIndex import SearchIndex
from peek_core_search._private.storage.SearchIndexCompilerQueue import \
    SearchIndexCompilerQueue

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.searchIndexCompilerQueueStatus = state
        self._adminStatusController.status.searchIndexCompilerQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.searchIndexCompilerQueueProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.searchIndexCompilerQueueLastError = error
        self._adminStatusController.notify()


class SearchIndexChunkCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 1.000

    # We can run this with a large queue because this compiler only runs when the
    # object compiler queue is empty
    QUEUE_BLOCKS_MAX = 40
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = SearchIndexCompilerQueue
    _VacuumDeclaratives = (SearchIndexCompilerQueue,
                           SearchIndex, EncodedSearchIndexChunk)

    def __init__(self, dbSessionCreator,
                 statusController: StatusController,
                 clientSearchIndexUpdateHandler: ClientSearchIndexChunkUpdateHandler,
                 isProcessorEnabledCallable: Callable):
        ACIProcessorQueueControllerABC \
            .__init__(self, dbSessionCreator,
                      _Notifier(statusController),
                      isProcessorEnabledCallable)

        self._clientSearchIndexUpdateHandler: ClientSearchIndexChunkUpdateHandler \
            = clientSearchIndexUpdateHandler

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_core_search._private.worker.tasks.SearchIndexChunkCompilerTask import \
            compileSearchIndexChunk

        return compileSearchIndexChunk.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientSearchIndexUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return '''
                 with sq_raw as (
                    SELECT "id", "chunkKey"
                    FROM core_search."SearchIndexCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "chunkKey"
                    FROM sq_raw
                    GROUP BY  "chunkKey"
                    HAVING count("chunkKey") > 1
                )
                DELETE
                FROM core_search."SearchIndexCompilerQueue"
                     USING sq sq1
                WHERE core_search."SearchIndexCompilerQueue"."id" != sq1."minId"
                    AND core_search."SearchIndexCompilerQueue"."id" > %(id)s
                    AND core_search."SearchIndexCompilerQueue"."chunkKey" = sq1."chunkKey"

            ''' % {'id': lastFetchedId, 'limit': dedupeLimit}



import logging
from typing import Optional

from vortex.rpc.RPC import vortexRPC

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkLoadRpcABC import \
    ACIChunkLoadRpcABC
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.storage.EncodedSearchIndexChunk import \
    EncodedSearchIndexChunk
from peek_core_search._private.storage.EncodedSearchObjectChunk import \
    EncodedSearchObjectChunk
from peek_plugin_base.PeekVortexUtil import peekServerName, peekClientName

logger = logging.getLogger(__name__)


class ClientChunkLoadRpc(ACIChunkLoadRpcABC):

    def makeHandlers(self):
        """ Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of it's self so we can simply yield the result
        of the start method.

        """

        yield self.loadSearchIndexChunks.start(funcSelf=self)
        yield self.loadSearchObjectChunks.start(funcSelf=self)
        logger.debug("RPCs started")

    # -------------
    @vortexRPC(peekServerName, acceptOnlyFromVortex=peekClientName, timeoutSeconds=60,
               additionalFilt=searchFilt, deferToThread=True)
    def loadSearchIndexChunks(self, offset: int, count: int) -> Optional[bytes]:
        """ Update Page Loader Status

        Tell the server of the latest status of the loader

        """
        return self.ckiInitialLoadChunksPayloadBlocking(offset, count,
                                                        EncodedSearchIndexChunk)

    # -------------
    @vortexRPC(peekServerName, acceptOnlyFromVortex=peekClientName, timeoutSeconds=60,
               additionalFilt=searchFilt, deferToThread=True)
    def loadSearchObjectChunks(self, offset: int, count: int) -> Optional[bytes]:
        """ Update Page Loader Status

        Tell the server of the latest status of the loader

        """
        return self.ckiInitialLoadChunksPayloadBlocking(offset, count,
                                                        EncodedSearchObjectChunk)



from vortex.handler.TupleDataObservableProxyHandler import TupleDataObservableProxyHandler

from peek_core_search._private.client.controller.SearchIndexCacheController import \
    SearchIndexCacheController
from peek_core_search._private.client.controller.SearchObjectCacheController import \
    SearchObjectCacheController
from peek_core_search._private.client.tuple_providers.ClientSearchIndexUpdateDateTupleProvider import \
    ClientSearchIndexUpdateDateTupleProvider
from peek_core_search._private.client.tuple_providers.ClientSearchObjectUpdateDateTupleProvider import \
    ClientSearchObjectUpdateDateTupleProvider
from peek_core_search._private.tuples.search_index.SearchIndexUpdateDateTuple import \
    SearchIndexUpdateDateTuple
from peek_core_search._private.tuples.search_object.SearchObjectUpdateDateTuple import \
    SearchObjectUpdateDateTuple


def makeClientTupleDataObservableHandler(
        tupleObservable: TupleDataObservableProxyHandler,
        searchIndexCacheHandler: SearchIndexCacheController,
        searchObjectCacheHandler: SearchObjectCacheController):
    """" Make CLIENT Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param tupleObservable: The tuple observable proxy
    :param searchIndexCacheHandler: The search index cache handler
    :param searchObjectCacheHandler: The search object cache handler
    :return: An instance of :code:`TupleDataObservableHandler`

    """

    tupleObservable.addTupleProvider(
        SearchIndexUpdateDateTuple.tupleName(),
        ClientSearchIndexUpdateDateTupleProvider(searchIndexCacheHandler))

    tupleObservable.addTupleProvider(
        SearchObjectUpdateDateTuple.tupleName(),
        ClientSearchObjectUpdateDateTupleProvider(searchObjectCacheHandler))



from twisted.internet.defer import inlineCallbacks
from twisted.trial import unittest

from peek_core_search._private.client.controller.FastKeywordController import \
    FastKeywordController
from peek_core_search._private.tuples.search_object.SearchResultObjectTuple import \
    SearchResultObjectTuple
from peek_core_search._private.worker.tasks.KeywordSplitter import splitPartialKeywords


class FastKeywordControllerTest(unittest.TestCase):
    def test_mergePartialAndFullMatches_1(self):
        fullByKw = {'^to$': [5, 6]}
        partialByKw = {'^to': [7], '^aa': [8]}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches('to', fullByKw, partialByKw)

        self.assertEqual(set(resultByKw), {'to'})
        self.assertEqual(len(resultByKw), 1)
        self.assertEqual(set(resultByKw['to']), {5, 6, 7})

    def test_mergePartialAndFullMatches_2(self):
        fullByKw = {'^five$': [5, 6]}
        partialByKw = {'^fiv': [7, 6],
                       'ive': [7, 6]}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches(
            "five seven", fullByKw, partialByKw)

        self.assertEqual(set(resultByKw), {'five'})
        self.assertEqual(len(resultByKw), 1)
        self.assertEqual(set(resultByKw['five']), {5, 6, 7})

    def test_mergePartialAndFullMatches_3(self):
        searchString = 'tatu west fus'
        fullByKw = {}
        partialByKw = {t: [7, 6] for t in splitPartialKeywords(searchString)}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches(
            searchString, fullByKw, partialByKw)

        self.assertEqual(set(resultByKw), {'west', 'fus', 'tatu'})
        self.assertEqual(set(resultByKw['west']), {6, 7})
        self.assertEqual(set(resultByKw['fus']), {6, 7})
        self.assertEqual(set(resultByKw['tatu']), {6, 7})

    def test_setIntersectFilterIndexResults(self):
        data = {'0': [6778979, 7042955]}

        inst = FastKeywordController(None, None)

        result = inst._setIntersectFilterIndexResults(data)

        self.assertEqual(result, {6778979, 7042955})
        self.assertEqual(len(result), len(data['0']))

    @inlineCallbacks
    def test_filterObjectsForSearchString(self):
        from twisted.python import threadable
        threadable.isInIOThread = lambda: True

        data = [SearchResultObjectTuple(properties={
            'any': 'TRANS HV FUSE TATURA WEST 28'
        })]

        inst = FastKeywordController(None, None)

        # No property name
        result = yield inst._filterObjectsForSearchString(
            data,
            'TRANS HV FUSE TATURA WEST 28',
            None)

        self.assertEqual(len(result), 1)

        # Now try it with the property name
        result = yield inst._filterObjectsForSearchString(
            data,
            'TRANS HV FUSE TATURA WEST 28',
            'any')

        self.assertEqual(len(result), 1)

        # Now try it with a WRONG the property name
        result = yield inst._filterObjectsForSearchString(
            data,
            'TRANS HV FUSE TATURA WEST 28',
            'dfgdfg')

        self.assertEqual(len(result), 0)

        # partial keywords search string
        result = yield inst._filterObjectsForSearchString(
            data,
            'tatu west fus',
            None)

        self.assertEqual(len(result), 1)



import logging
from typing import Dict

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import \
    ACICacheHandlerABC
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import \
    ACIUpdateDateTupleABC
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.tuples.search_object.SearchObjectUpdateDateTuple import \
    SearchObjectUpdateDateTuple

logger = logging.getLogger(__name__)

clientSearchObjectWatchUpdateFromDeviceFilt = {
    'key': "clientSearchObjectWatchUpdateFromDevice"
}
clientSearchObjectWatchUpdateFromDeviceFilt.update(searchFilt)


# ModelSet HANDLER
class SearchObjectCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = SearchObjectUpdateDateTuple
    _updateFromServerFilt: Dict = clientSearchObjectWatchUpdateFromDeviceFilt
    _logger: logging.Logger = logger



"""added SI objectId index

Peek Plugin Database Migration Script

Revision ID: 2c6cad1f280e
Revises: 30f0d0f925ca
Create Date: 2018-07-01 18:03:05.530637

"""

# revision identifiers, used by Alembic.
revision = '2c6cad1f280e'
down_revision = '30f0d0f925ca'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_SearchIndex_objectId', 'SearchIndex', ['objectId'], unique=False, schema='core_search')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_SearchIndex_objectId', table_name='SearchIndex', schema='core_search')
    # ### end Alembic commands ###


import logging

from sqlalchemy import Column, BigInteger
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index, ForeignKey

from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class SearchObject(Tuple, DeclarativeBase):
    __tablename__ = 'SearchObject'
    __tupleType__ = searchTuplePrefix + 'SearchObjectTable'

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    #:  The object that this routs is for
    objectTypeId = Column(Integer,
                          ForeignKey('SearchObjectType.id', ondelete='CASCADE'),
                          nullable=False)

    key = Column(String, nullable=False)

    chunkKey = Column(Integer, nullable=False)

    fullKwPropertiesJson = Column(String, nullable=True)

    partialKwPropertiesJson = Column(String, nullable=True)

    packedJson = Column(String, nullable=True)

    __table_args__ = (
        Index("idx_SearchObject_objectTypeId", objectTypeId),
        Index("idx_SearchObject_key", key, unique=True),
        Index("idx_SearchObject_chunkKey", chunkKey),
    )



import logging

from peek_core_search._private.PluginNames import searchTuplePrefix
from sqlalchemy import Column, BigInteger
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index, ForeignKey

from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)



@addTupleType
class SearchIndex(Tuple, DeclarativeBase):
    __tablename__ = 'SearchIndex'
    __tupleType__ = searchTuplePrefix + 'SearchIndexTable'

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    chunkKey = Column(Integer, nullable=False)
    keyword = Column(String, nullable=False)
    propertyName = Column(String, nullable=False)

    #:  The object that this routs is for
    objectId = Column(BigInteger,
                      ForeignKey('SearchObject.id', ondelete='CASCADE'),
                      nullable=False)

    __table_args__ = (
        Index("idx_SearchIndex_quick_query",
              chunkKey, keyword, propertyName, objectId,
              unique=True),
        Index("idx_SearchIndex_objectId", objectId), # For foreign key
    )



from sqlalchemy import Column, Index, Boolean
from sqlalchemy import Integer, String

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType


@addTupleType
class SearchPropertyTuple(Tuple, DeclarativeBase):
    __tupleType__ = searchTuplePrefix + 'SearchPropertyTuple'
    __tablename__ = 'SearchProperty'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    order = Column(Integer, nullable=False, server_default='0')

    showOnResult = Column(Boolean, nullable=False, server_default='1')
    showInHeader = Column(Boolean, nullable=False, server_default='0')

    __table_args__ = (
        Index("idx_SearchProp_name", name, unique=True),
        Index("idx_SearchProp_title", title, unique=True),
    )



from twisted.trial import unittest

from peek_core_search._private.worker.tasks.KeywordSplitter import \
    splitPartialKeywords, splitFullKeywords, _splitFullTokens, twoCharTokens


class KeywordSplitterTest(unittest.TestCase):
    def testDuplicates(self):
        self.assertEqual({'smith'},
                         _splitFullTokens("smith smith"))

    def test_twoCharTokens(self):
        self.assertEqual(set(),
                         twoCharTokens("smith smith"))

        self.assertEqual({'^to$'},
                         twoCharTokens(splitFullKeywords("two to")))

        self.assertEqual({'^to'},
                         twoCharTokens(splitPartialKeywords("two to")))

    def testFullKeywordSplit(self):
        self.assertEqual({'^smith$'},
                         splitFullKeywords("smith"))
        self.assertEqual({'^zorroreyner$'},
                         splitFullKeywords("ZORRO-REYNER"))
        self.assertEqual({'^34534535$'},
                         splitFullKeywords("34534535"))

        self.assertEqual({'^and$'},
                         splitFullKeywords("and"))
        self.assertEqual({'^to$'},
                         splitFullKeywords("to"))
        self.assertEqual({'^to$'},
                         splitFullKeywords("to"))

        self.assertEqual({'^milton$', '^unit$', '^22$'},
                         splitFullKeywords("Milton Unit 22"))

        self.assertEqual({'^unit$'},
                         splitFullKeywords("Unit A"))

        self.assertEqual({'^unit$'},
                         splitFullKeywords("Unit 1"))

        self.assertEqual({'^trans$', '^66kv$', '^b3$', '^cb$', '^ats$'},
                         splitFullKeywords("ATS B3 TRANS 66KV CB"))

    def testPartialKeywordSplit(self):
        self.assertEqual({'^smi', 'mit', 'ith'},
                         splitPartialKeywords("smith"))

        self.assertEqual(
            {'^zor', 'orr', 'rro', 'ror', 'ore', 'rey', 'eyn', 'yne', 'ner'},
            splitPartialKeywords("ZORRO-REYNER"))
        self.assertEqual({'^345', '535', '453', '534', '345'},
                         splitPartialKeywords("34534535"))

        self.assertEqual({'^and'},
                         splitPartialKeywords("and"))

        self.assertEqual({"^to"},
                         splitPartialKeywords("to"))

        self.assertEqual({'a55', '^ha5'},
                         splitPartialKeywords("ha55"))

        self.assertEqual({'^mil', 'ilt', 'lto', 'ton', '^uni', 'nit', '^22'},
                         splitPartialKeywords("Milton Unit 22"))

        self.assertEqual({'^mil', 'ill', 'lls', '^un', '^no'},
                         splitPartialKeywords("mills un no"))

        self.assertEqual({'^uni', 'nit', "^22"},
                         splitPartialKeywords("Unit 22"))

        self.assertEqual({'^uni', 'nit'},
                         splitPartialKeywords("Unit 1"))

        self.assertEqual({'^uni', 'nit'},
                         splitPartialKeywords("A Unit"))

        self.assertEqual({'^uni', 'nit'},
                         splitPartialKeywords("2 Unit"))

        self.assertEqual({'^ats', "^cb", "^b3",
                          '^66k', '6kv', '^tra', 'ran', 'ans'},
                         splitPartialKeywords("ATS B3 TRANS 66KV CB"))

        self.assertEqual({'^col', '^lin', 'ins'},
                         splitPartialKeywords("COL LINS"))

        self.assertNotEqual(splitPartialKeywords("COLLINS"),
                            splitPartialKeywords("COL LINS"))



"""added show to prop

Peek Plugin Database Migration Script

Revision ID: 36aeb6eaea5c
Revises: c18d0a2ab678
Create Date: 2019-07-30 21:10:50.837891

"""

# revision identifiers, used by Alembic.
revision = '36aeb6eaea5c'
down_revision = 'c18d0a2ab678'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ###
    op.add_column('SearchProperty', sa.Column('showInHeader', sa.Boolean(), server_default='0', nullable=True), schema='core_search')


    op.execute(''' UPDATE "core_search"."SearchProperty" SET "showInHeader" = false ''')

    op.alter_column('SearchProperty', 'showInHeader',
                    type_=sa.Boolean(), nullable=False,
                    schema='core_search')

    # ###
    op.add_column('SearchProperty', sa.Column('showOnResult', sa.Boolean(), server_default='1', nullable=True), schema='core_search')

    op.execute(''' UPDATE "core_search"."SearchProperty" SET "showOnResult" = true ''')

    op.alter_column('SearchProperty', 'showOnResult',
                    type_=sa.Boolean(), nullable=False,
                    schema='core_search')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('SearchProperty', 'showOnResult', schema='core_search')
    op.drop_column('SearchProperty', 'showInHeader', schema='core_search')
    # ### end Alembic commands ###


from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class AdminStatusTuple(Tuple):
    __tupleType__ = searchTuplePrefix + "AdminStatusTuple"

    searchIndexCompilerQueueStatus: bool = TupleField(False)
    searchIndexCompilerQueueSize: int = TupleField(0)
    searchIndexCompilerQueueProcessedTotal: int = TupleField(0)
    searchIndexCompilerQueueLastError: str = TupleField()


    searchObjectCompilerQueueStatus: bool = TupleField(False)
    searchObjectCompilerQueueSize: int = TupleField(0)
    searchObjectCompilerQueueProcessedTotal: int = TupleField(0)
    searchObjectCompilerQueueLastError: str = TupleField()



import logging

logger = logging.getLogger(__name__)

INDEX_BUCKET_COUNT = 8192
OBJECT_BUCKET_COUNT = 8192


def makeSearchIndexChunkKey(key: str) -> int:
    """ Make Chunk Key

    This is simple, and provides a reasonable distribution

    :param key:

    :return: chunkKey

    """

    if not key:
        raise Exception("key is None or zero length")

    bucket = 0
    for char in key:
        bucket = ((bucket << 5) - bucket) + ord(char)
        bucket = bucket | 0  # This is in the javascript code.

    bucket = bucket & (INDEX_BUCKET_COUNT - 1)

    return bucket


def makeSearchObjectChunkKey(key: int) -> int:
    """ Make Chunk Key

    This is simple, and provides a reasonable distribution

    :param key:

    :return: chunkKey

    """

    if key is None:
        raise Exception("key is None")

    bucket = key & (OBJECT_BUCKET_COUNT - 1)

    return bucket



import logging
from typing import Union

from twisted.internet.defer import Deferred

from peek_core_search._private.storage.SearchPropertyTuple import SearchPropertyTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class SearchPropertyTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:

        session = self._ormSessionCreator()
        try:
            tuples = (
                session
                    .query(SearchPropertyTuple)
                    .order_by(SearchPropertyTuple.title)
                    .all()
            )

            # Create the vortex message
            return Payload(filt, tuples=tuples).makePayloadEnvelope().toVortexMsg()

        finally:
            session.close()



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchActionProcessorName
from .controller.MainController import MainController


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=searchActionProcessorName,
        additionalFilt=searchFilt,
        defaultDelegate=mainController)
    return processor



from peek_core_search._private.client.controller.FastKeywordController import \
    FastKeywordController
from peek_core_search._private.tuples.KeywordAutoCompleteTupleAction import \
    KeywordAutoCompleteTupleAction
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchActionProcessorName
from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy


def makeTupleActionProcessorProxy(fastKeywordController: FastKeywordController):
    proxy = TupleActionProcessorProxy(
        tupleActionProcessorName=searchActionProcessorName,
        proxyToVortexName=peekServerName,
        additionalFilt=searchFilt)

    proxy.setDelegate(KeywordAutoCompleteTupleAction.tupleType(),
                      fastKeywordController)

    return proxy



from typing import List

from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_core_search._private.PluginNames import searchTuplePrefix
from .SearchResultDetailTuple import SearchResultDetailTuple


@addTupleType
class SearchResultTuple(Tuple):
    """ Search Result Tuple

    This tuple represents a search result
    """
    __tupleType__ = searchTuplePrefix + 'SearchResultTuple'

    #:  The id of the object this search result is for
    objectId: str = TupleField()

    #:  The type of this object in the search result
    objectType: str = TupleField()

    #:  The details of the search result
    details: List[SearchResultDetailTuple] = TupleField(defaultValue=[])



from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class ImportSearchObjectRouteTuple(Tuple):
    """ Import Search Object Route

    This tuple is used to import object routes into the search plugin

    """
    __tupleType__ = searchTuplePrefix + 'ImportSearchObjectRouteTuple'

    #:  A unique string describing this group being imported
    # This can be used for deleting as well
    importGroupHash: str = TupleField()

    #:  The title of the route, that the user will see
    routeTitle: str = TupleField()

    #:  The route that the angular app will route to
    routePath: str = TupleField()
